@@ -1,5 +1,5 @@
 /* libunwind - a platform-independent unwind library
-   Copyright (C) 2001-2003 Hewlett-Packard Co
+   Copyright (C) 2001-2004 Hewlett-Packard Co
 	Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
 
 This file is part of libunwind.
@@ -26,14 +26,13 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #include <alloca.h>
 #include <stdlib.h>
 
-#include "rse.h"
 #include "unwind_i.h"
 
 #ifdef HAVE_SYS_UC_ACCESS_H
 # include <sys/uc_access.h>
 #endif
 
-#ifdef UNW_DEBUG
+#if UNW_DEBUG
 
 HIDDEN const char *
 ia64_strloc (ia64_loc_t loc)
@@ -73,72 +72,24 @@ ia64_strloc (ia64_loc_t loc)
 #ifdef UNW_REMOTE_ONLY
 
 /* unw_local_addr_space is a NULL pointer in this case.  */
-unw_addr_space_t unw_local_addr_space;
+PROTECTED unw_addr_space_t unw_local_addr_space;
 
 #else /* !UNW_REMOTE_ONLY */
 
 static struct unw_addr_space local_addr_space;
 
-unw_addr_space_t unw_local_addr_space = &local_addr_space;
+PROTECTED unw_addr_space_t unw_local_addr_space = &local_addr_space;
 
 #ifdef HAVE_SYS_UC_ACCESS_H
 
 #else /* !HAVE_SYS_UC_ACCESS_H */
 
-static inline void *
-uc_addr (ucontext_t *uc, int reg)
+HIDDEN void *
+tdep_uc_addr (ucontext_t *uc, int reg, uint8_t *nat_bitnr)
 {
-  void *addr;
-
-  switch (reg)
-    {
-    case UNW_IA64_GR + 0:	addr = &unw.r0; break;
-    case UNW_IA64_NAT + 0:	addr = &unw.r0; break;
-    case UNW_IA64_IP:		addr = &uc->uc_mcontext.sc_br[0]; break;
-    case UNW_IA64_CFM:		addr = &uc->uc_mcontext.sc_ar_pfs; break;
-    case UNW_IA64_AR_RNAT:	addr = &uc->uc_mcontext.sc_ar_rnat; break;
-    case UNW_IA64_AR_UNAT:	addr = &uc->uc_mcontext.sc_ar_unat; break;
-    case UNW_IA64_AR_LC:	addr = &uc->uc_mcontext.sc_ar_lc; break;
-    case UNW_IA64_AR_FPSR:	addr = &uc->uc_mcontext.sc_ar_fpsr; break;
-    case UNW_IA64_PR:		addr = &uc->uc_mcontext.sc_pr; break;
-      /* This may look confusing, but it's correct: AR_BSPSTORE needs
-         to return the address past the last word written, which is
-         stored in sc_ar_bsp.  On the other hand, AR_BSP needs to
-         return the address that was in ar.bsp at the time the context
-         was captured.  As described in unw_init_local(), sc_ar_bsp is
-         (ab-)used for this purpose.  */
-    case UNW_IA64_AR_BSP:	addr = &uc->uc_mcontext.sc_rbs_base; break;
-    case UNW_IA64_AR_BSPSTORE:	addr = &uc->uc_mcontext.sc_ar_bsp; break;
-
-    case UNW_IA64_GR + 4 ... UNW_IA64_GR + 7:
-    case UNW_IA64_GR + 12:
-      addr = &uc->uc_mcontext.sc_gr[reg - UNW_IA64_GR];
-      break;
-
-    case UNW_IA64_BR + 1 ... UNW_IA64_BR + 5:
-      addr = &uc->uc_mcontext.sc_br[reg - UNW_IA64_BR];
-      break;
-
-    case UNW_IA64_FR+ 2 ... UNW_IA64_FR+ 5:
-    case UNW_IA64_FR+16 ... UNW_IA64_FR+31:
-      addr = &uc->uc_mcontext.sc_fr[reg - UNW_IA64_FR];
-      break;
-
-    default:
-      addr = NULL;
-    }
-  return addr;
+  return inlined_uc_addr (uc, reg, nat_bitnr);
 }
 
-# ifdef UNW_LOCAL_ONLY
-
-void *
-tdep_uc_addr (ucontext_t *uc, int reg)
-{
-  return uc_addr (uc, reg);
-}
-
-# endif /* UNW_LOCAL_ONLY */
 #endif /* !HAVE_SYS_UC_ACCESS_H */
 
 static void
@@ -151,7 +102,7 @@ static int
 get_dyn_info_list_addr (unw_addr_space_t as, unw_word_t *dyn_info_list_addr,
 			void *arg)
 {
-  *dyn_info_list_addr = (unw_word_t) &_U_dyn_info_list;
+  *dyn_info_list_addr = _U_dyn_info_list_addr ();
   return 0;
 }
 
@@ -161,13 +112,13 @@ access_mem (unw_addr_space_t as, unw_word_t addr, unw_word_t *val, int write,
 {
   if (write)
     {
-      debug (100, "%s: mem[%lx] <- %lx\n", __FUNCTION__, addr, *val);
+      Debug (12, "mem[%lx] <- %lx\n", addr, *val);
       *(unw_word_t *) addr = *val;
     }
   else
     {
       *val = *(unw_word_t *) addr;
-      debug (100, "%s: mem[%lx] -> %lx\n", __FUNCTION__, addr, *val);
+      Debug (12, "mem[%lx] -> %lx\n", addr, *val);
     }
   return 0;
 }
@@ -285,15 +236,15 @@ access_reg (unw_addr_space_t as, unw_regnum_t reg, unw_word_t *val, int write,
 
   if (ret != 0)
     {
-      debug (1, "%s: failed to %s %s\n",
-	     __FUNCTION__, write ? "write" : "read", unw_regname (reg));
+      Debug (1, "failed to %s %s (ret = %d)\n",
+	     write ? "write" : "read", unw_regname (reg), ret);
       return -UNW_EBADREG;
     }
 
   if (write)
-    debug (100, "%s: %s <- %lx\n", __FUNCTION__, unw_regname (reg), *val);
+    Debug (12, "%s <- %lx\n", unw_regname (reg), *val);
   else
-    debug (100, "%s: %s -> %lx\n", __FUNCTION__, unw_regname (reg), *val);
+    Debug (12, "%s -> %lx\n", unw_regname (reg), *val);
   return 0;
 }
 
@@ -339,9 +290,6 @@ access_reg (unw_addr_space_t as, unw_regnum_t reg, unw_word_t *val, int write,
   unw_word_t *addr, mask;
   ucontext_t *uc = arg;
 
-  if (reg >= UNW_IA64_FR && reg < UNW_IA64_FR + 128)
-    goto badreg;
-
   if (reg >= UNW_IA64_NAT + 4 && reg <= UNW_IA64_NAT + 7)
     {
       mask = ((unw_word_t) 1) << (reg - UNW_IA64_NAT);
@@ -356,30 +304,35 @@ access_reg (unw_addr_space_t as, unw_regnum_t reg, unw_word_t *val, int write,
 	*val = (uc->uc_mcontext.sc_nat & mask) != 0;
 
       if (write)
-	debug (100, "%s: %s <- %lx\n", __FUNCTION__, unw_regname (reg), *val);
+	Debug (12, "%s <- %lx\n", unw_regname (reg), *val);
       else
-	debug (100, "%s: %s -> %lx\n", __FUNCTION__, unw_regname (reg), *val);
+	Debug (12, "%s -> %lx\n", unw_regname (reg), *val);
       return 0;
     }
 
-  addr = uc_addr (uc, reg);
+  addr = tdep_uc_addr (uc, reg, NULL);
   if (!addr)
     goto badreg;
 
   if (write)
     {
-      *(unw_word_t *) addr = *val;
-      debug (100, "%s: %s <- %lx\n", __FUNCTION__, unw_regname (reg), *val);
+      if (ia64_read_only_reg (addr))
+	{
+	  Debug (16, "attempt to write read-only register\n");
+	  return -UNW_EREADONLYREG;
+	}
+      *addr = *val;
+      Debug (12, "%s <- %lx\n", unw_regname (reg), *val);
     }
   else
     {
       *val = *(unw_word_t *) addr;
-      debug (100, "%s: %s -> %lx\n", __FUNCTION__, unw_regname (reg), *val);
+      Debug (12, "%s -> %lx\n", unw_regname (reg), *val);
     }
   return 0;
 
  badreg:
-  debug (1, "%s: bad register number %u\n", __FUNCTION__, reg);
+  Debug (1, "bad register number %u\n", reg);
   return -UNW_EBADREG;
 }
 
@@ -393,26 +346,31 @@ access_fpreg (unw_addr_space_t as, unw_regnum_t reg, unw_fpreg_t *val,
   if (reg < UNW_IA64_FR || reg >= UNW_IA64_FR + 128)
     goto badreg;
 
-  addr = uc_addr (uc, reg);
+  addr = tdep_uc_addr (uc, reg, NULL);
   if (!addr)
     goto badreg;
 
   if (write)
     {
-      debug (100, "%s: %s <- %016lx.%016lx\n", __FUNCTION__,
+      if (ia64_read_only_reg (addr))
+	{
+	  Debug (16, "attempt to write read-only register\n");
+	  return -UNW_EREADONLYREG;
+	}
+      *addr = *val;
+      Debug (12, "%s <- %016lx.%016lx\n",
 	     unw_regname (reg), val->raw.bits[1], val->raw.bits[0]);
-      *(unw_fpreg_t *) addr = *val;
     }
   else
     {
       *val = *(unw_fpreg_t *) addr;
-      debug (100, "%s: %s -> %016lx.%016lx\n", __FUNCTION__,
+      Debug (12, "%s -> %016lx.%016lx\n",
 	     unw_regname (reg), val->raw.bits[1], val->raw.bits[0]);
     }
   return 0;
 
  badreg:
-  debug (1, "%s: bad register number %u\n", __FUNCTION__, reg);
+  Debug (1, "bad register number %u\n", reg);
   /* attempt to access a non-preserved register */
   return -UNW_EBADREG;
 }
@@ -431,7 +389,6 @@ HIDDEN void
 ia64_local_addr_space_init (void)
 {
   memset (&local_addr_space, 0, sizeof (local_addr_space));
-  ia64_script_cache_init (&local_addr_space.global_cache);
   local_addr_space.big_endian = (__BYTE_ORDER == __BIG_ENDIAN);
 #if defined(__linux)
   local_addr_space.abi = ABI_LINUX;
@@ -439,7 +396,7 @@ ia64_local_addr_space_init (void)
   local_addr_space.abi = ABI_HPUX;
 #endif
   local_addr_space.caching_policy = UNW_CACHE_GLOBAL;
-  local_addr_space.acc.find_proc_info = UNW_ARCH_OBJ (find_proc_info);
+  local_addr_space.acc.find_proc_info = tdep_find_proc_info;
   local_addr_space.acc.put_unwind_info = put_unwind_info;
   local_addr_space.acc.get_dyn_info_list_addr = get_dyn_info_list_addr;
   local_addr_space.acc.access_mem = access_mem;
@@ -447,10 +404,13 @@ ia64_local_addr_space_init (void)
   local_addr_space.acc.access_fpreg = access_fpreg;
   local_addr_space.acc.resume = ia64_local_resume;
   local_addr_space.acc.get_proc_name = get_static_proc_name;
+  unw_flush_cache (&local_addr_space, 0, 0);
 }
 
 #endif /* !UNW_REMOTE_ONLY */
 
+#ifndef UNW_LOCAL_ONLY
+
 HIDDEN int
 ia64_uc_access_reg (struct cursor *c, ia64_loc_t loc, unw_word_t *valp,
 		    int write)
@@ -460,7 +420,7 @@ ia64_uc_access_reg (struct cursor *c, ia64_loc_t loc, unw_word_t *valp,
   ucontext_t *ucp;
   int ret;
 
-  debug (100, "%s: %s locaction %s\n", __FUNCTION__,
+  Debug (16, "%s location %s\n",
 	 write ? "writing" : "reading", ia64_strloc (loc));
 
   if (c->as == unw_local_addr_space)
@@ -575,3 +535,5 @@ ia64_uc_access_fpreg (struct cursor *c, ia64_loc_t loc, unw_fpreg_t *valp,
   return -UNW_EINVAL;
 #endif /* !HAVE_SYS_UC_ACCESS_H */
 }
+
+#endif /* UNW_LOCAL_ONLY */
