@@ -27,30 +27,28 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #include <stdlib.h>
 
 #include "init.h"
-#include "rse.h"
 #include "unwind_i.h"
 
 #ifdef UNW_REMOTE_ONLY
 
-int
-unw_init_local (unw_cursor_t *cursor, ucontext_t *uc)
+PROTECTED int
+unw_init_local (unw_cursor_t *cursor, unw_context_t *uc)
 {
   return -UNW_EINVAL;
 }
 
 #else /* !UNW_REMOTE_ONLY */
 
-int
-unw_init_local (unw_cursor_t *cursor, ucontext_t *uc)
+PROTECTED int
+unw_init_local (unw_cursor_t *cursor, unw_context_t *uc)
 {
   struct cursor *c = (struct cursor *) cursor;
   unw_word_t sol;
 
-  if (unw.needs_initialization)
-    {
-      unw.needs_initialization = 0;
-      ia64_init ();
-    }
+  if (tdep_needs_initialization)
+    tdep_init ();
+
+  Debug (2, "(cursor=%p)\n", c);
 
 #ifdef __hpux
   {
