@@ -1,5 +1,5 @@
 /* libunwind - a platform-independent unwind library
-   Copyright (C) 2001-2003 Hewlett-Packard Co
+   Copyright (C) 2001-2004 Hewlett-Packard Co
 	Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
 
 This file is part of libunwind.
@@ -52,7 +52,7 @@ push (struct ia64_state_record *sr)
   rs = alloc_reg_state ();
   if (!rs)
     {
-      fprintf (stderr, "libunwind: cannot stack reg state!\n");
+      print_error ("libunwind: cannot stack reg state!\n");
       return;
     }
   memcpy (rs, &sr->curr, sizeof (*rs));
@@ -66,7 +66,7 @@ pop (struct ia64_state_record *sr)
 
   if (!rs)
     {
-      fprintf (stderr, "libunwind: stack underflow!\n");
+      print_error ("libunwind: stack underflow!\n");
       return;
     }
   memcpy (&sr->curr, rs, sizeof (*rs));
@@ -84,7 +84,7 @@ dup_state_stack (struct ia64_reg_state *rs)
       copy = alloc_reg_state ();
       if (!copy)
 	{
-	  fprintf (stderr, "unwind.dup_state_stack: out of memory\n");
+	  print_error ("unwind.dup_state_stack: out of memory\n");
 	  return NULL;
 	}
       memcpy (copy, rs, sizeof (*copy));
@@ -522,7 +522,7 @@ desc_copy_state (unw_word label, struct ia64_state_record *sr)
 	  return;
 	}
     }
-  fprintf (stderr, "libunwind: failed to find state labeled 0x%lx\n", label);
+  print_error ("libunwind: failed to find labeled state\n");
 }
 
 static inline void
@@ -533,7 +533,7 @@ desc_label_state (unw_word label, struct ia64_state_record *sr)
   ls = alloc_labeled_state ();
   if (!ls)
     {
-      fprintf (stderr, "unwind.desc_label_state(): out of memory\n");
+      print_error ("unwind.desc_label_state(): out of memory\n");
       return;
     }
   ls->label = label;
@@ -629,7 +629,7 @@ desc_spill_sprel_p (unsigned char qp, unw_word t, unsigned char abreg,
 }
 
 #define UNW_DEC_BAD_CODE(code)						\
-	fprintf (stderr, "libunwind: unknown code 0x%02x\n", code)
+	print_error ("libunwind: unknown code encountered\n")
 
 /* Register names.  */
 #define UNW_REG_BSP		IA64_REG_BSP
@@ -744,6 +744,48 @@ lookup_preg (int regnum, int memory, struct ia64_state_record *sr)
   return sr->curr.reg + preg;
 }
 
+/* An alias directive inside a region of length RLEN is interpreted to
+   mean that the region behaves exactly like the first RLEN
+   instructions at the aliased IP.  RLEN=0 implies that the current
+   state matches exactly that of the aliased IP.  */
+
+static int
+desc_alias (unw_dyn_op_t *op, struct cursor *c, struct ia64_state_record *sr)
+{
+  struct ia64_state_record orig_sr = *sr;
+  int i, ret, when, rlen = sr->region_len;
+  unw_word_t new_ip;
+
+  when = MIN(sr->when_target, rlen - 1);
+  new_ip = op->val + ((when / 3) * 16 + (when % 3));
+
+  if ((ret = ia64_fetch_proc_info (c, new_ip, 1)) < 0)
+    return ret;
+
+  if ((ret = create_state_record_for (c, sr, new_ip)) < 0)
+    return ret;
+
+  sr->first_region = orig_sr.first_region;
+  sr->done = 0;
+  sr->any_spills |= orig_sr.any_spills;
+  sr->in_body = orig_sr.in_body;
+  sr->region_start = orig_sr.region_start;
+  sr->region_len = orig_sr.region_len;
+  if (sr->when_sp_restored != IA64_WHEN_NEVER)
+    sr->when_sp_restored = op->when + MIN (orig_sr.when_sp_restored, rlen - 1);
+  sr->epilogue_count = orig_sr.epilogue_count;
+  sr->when_target = orig_sr.when_target;
+
+  for (i = 0; i < IA64_NUM_PREGS; ++i)
+    if (sr->curr.reg[i].when != IA64_WHEN_NEVER)
+      sr->curr.reg[i].when = op->when + MIN (sr->curr.reg[i].when, rlen - 1);
+
+  ia64_free_state_record (sr);
+  sr->labeled_states = orig_sr.labeled_states;
+  sr->curr.next = orig_sr.curr.next;
+  return 0;
+}
+
 static inline int
 parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 {
@@ -751,10 +793,10 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
   unw_dyn_proc_info_t *proc = &di->u.pi;
   unw_dyn_region_info_t *r;
   struct ia64_reg_info *ri;
-  unw_word_t val, new_ip;
   enum ia64_where where;
-  unw_dyn_op_t *op;
   int32_t when, len;
+  unw_dyn_op_t *op;
+  unw_word_t val;
   int memory, ret;
   int8_t qp;
 
@@ -765,8 +807,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 	{
 	  if (r->next)
 	    {
-	      debug (10, "libunwind: negative region length allowed in last "
-		     "region only!");
+	      Debug (1, "negative region length allowed in last region only!");
 	      return -UNW_EINVAL;
 	    }
 	  len = -len;
@@ -862,14 +903,8 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 	      break;
 
 	    case UNW_DYN_ALIAS:
-	      while (sr->curr.next)
-		pop (sr);
-	      new_ip = op->val + ((sr->when_target / 3) * 16
-				  + (sr->when_target % 3));
-	      ret = ia64_fetch_proc_info (c, new_ip, 1);
-	      if (ret < 0)
+	      if ((ret = desc_alias (op, c, sr)) < 0)
 		return ret;
-	      return create_state_record_for (c, sr, new_ip);
 
 	    case UNW_DYN_STOP:
 	      goto end_of_ops;
@@ -910,10 +945,10 @@ put_unwind_info (struct cursor *c, unw_proc_info_t *pi)
   if (!c->pi_valid)
     return;
 
-  if ((c)->pi_is_dynamic)
+  if (c->pi_is_dynamic)
     unwi_put_dynamic_unwind_info (c->as, pi, c->as_arg);
   else
-    ia64_put_unwind_info(c, pi);
+    ia64_put_unwind_info (c, pi);
 }
 
 static int
@@ -933,13 +968,12 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
     r->when = IA64_WHEN_NEVER;
   sr->pr_val = predicates;
   sr->first_region = 1;
-  sr->return_link_reg = 0;
 
   if (!c->pi.unwind_info)
     {
       /* No info, return default unwinder (leaf proc, no mem stack, no
          saved regs), rp in b0, pfs in ar.pfs.  */
-      debug (1, "unwind.parser: no unwind info for ip=0x%lx (gp=%lx)\n",
+      Debug (1, "no unwind info for ip=0x%lx (gp=%lx)\n",
 	     (long) ip, (long) c->pi.gp);
       sr->curr.reg[IA64_REG_IP].where = IA64_WHERE_BR;
       sr->curr.reg[IA64_REG_IP].when = -1;
@@ -950,16 +984,24 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
   sr->when_target = (3 * ((ip & ~(unw_word_t) 0xf) - c->pi.start_ip) / 16
 		     + (ip & 0xf));
 
-  if (c->pi.format == UNW_INFO_FORMAT_TABLE)
+  switch (c->pi.format)
     {
+    case UNW_INFO_FORMAT_TABLE:
+    case UNW_INFO_FORMAT_REMOTE_TABLE:
       dp = c->pi.unwind_info;
       desc_end = dp + c->pi.unwind_info_size;
       while (!sr->done && dp < desc_end)
 	dp = unw_decode (dp, sr->in_body, sr);
       ret = 0;
+      break;
+
+    case UNW_INFO_FORMAT_DYNAMIC:
+      ret = parse_dynamic (c, sr);
+      break;
+
+    default:
+      ret = -UNW_EINVAL;
     }
-  else
-    ret = parse_dynamic (c, sr);
 
   put_unwind_info (c, &c->pi);
 
@@ -996,14 +1038,13 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
       && sr->when_target > sr->curr.reg[IA64_REG_BSPSTORE].when
       && sr->when_target > sr->curr.reg[IA64_REG_RNAT].when)
     {
-      debug (10,
-	     "libunwind: func 0x%lx may switch the register-backing-store\n",
+      Debug (8, "func 0x%lx may switch the register-backing-store\n",
 	     c->pi.start_ip);
       c->pi.flags |= UNW_PI_FLAG_IA64_RBS_SWITCH;
     }
  out:
 #if UNW_DEBUG
-  if (unw.debug_level > 0)
+  if (unwi_debug_level > 0)
     {
       dprintf ("libunwind: state record for func 0x%lx, t=%u (flags=0x%lx):\n",
 	       (long) c->pi.start_ip, sr->when_target, (long) c->pi.flags);
@@ -1050,7 +1091,7 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
 HIDDEN int
 ia64_create_state_record (struct cursor *c, struct ia64_state_record *sr)
 {
-  return create_state_record_for(c, sr, c->ip);
+  return create_state_record_for (c, sr, c->ip);
 }
 
 HIDDEN int
