@@ -1,5 +1,5 @@
 /* libunwind - a platform-independent unwind library
-   Copyright (C) 2001-2003 Hewlett-Packard Co
+   Copyright (C) 2001-2004 Hewlett-Packard Co
 	Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
 
 This file is part of libunwind.
@@ -24,7 +24,6 @@ OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 
 #include "offsets.h"
-#include "rse.h"
 #include "unwind_i.h"
 
 static inline int
@@ -218,7 +217,7 @@ check_rbs_switch (struct cursor *c)
 static inline int
 update_frame_state (struct cursor *c)
 {
-  unw_word_t prev_ip, prev_sp, prev_bsp, ip, pr, num_regs;
+  unw_word_t prev_ip, prev_sp, prev_bsp, ip, num_regs;
   int ret;
 
   prev_ip = c->ip;
@@ -236,7 +235,7 @@ update_frame_state (struct cursor *c)
   if ((ip & 0xc) != 0)
     {
       /* don't let obviously bad addresses pollute the cache */
-      debug (1, "%s: rejecting bad ip=0x%lx\n",  __FUNCTION__, (long) c->ip);
+      Debug (1, "rejecting bad ip=0x%lx\n", (long) c->ip);
       return -UNW_EINVALIDIP;
     }
   if (ip == 0)
@@ -276,12 +275,12 @@ update_frame_state (struct cursor *c)
 	  break;
 
 	default:
-	  debug (1, "%s: unknown ABI marker: ABI=%u, context=%u\n",
-		 __FUNCTION__, c->abi_marker >> 8, c->abi_marker & 0xff);
+	  Debug (1, "unknown ABI marker: ABI=%u, context=%u\n",
+		 c->abi_marker >> 8, c->abi_marker & 0xff);
 	  return -UNW_EINVAL;
 	}
-      debug (100, "%s: sigcontext_addr=%lx (ret=%d)\n",
-	     __FUNCTION__, (unsigned long) c->sigcontext_addr, ret);
+      Debug (12, "sigcontext_addr=%lx (ret=%d)\n",
+	     (unsigned long) c->sigcontext_addr, ret);
 
       c->sigcontext_off = c->sigcontext_addr - c->sp;
 
@@ -305,7 +304,6 @@ update_frame_state (struct cursor *c)
 
   c->bsp = ia64_rse_skip_regs (c->bsp, -num_regs);
 
-  pr = c->pr;
   c->sp = c->psp;
   c->abi_marker = 0;
 
@@ -329,12 +327,14 @@ update_frame_state (struct cursor *c)
 }
 
 
-int
+PROTECTED int
 unw_step (unw_cursor_t *cursor)
 {
   struct cursor *c = (struct cursor *) cursor;
   int ret;
 
+  Debug (2, "(cursor=%p)\n", c);
+
   ret = ia64_find_save_locs (c);
   if (ret < 0)
     return ret;
