@@ -1,5 +1,5 @@
 /* libunwind - a platform-independent unwind library
-   Copyright (C) 2001-2003 Hewlett-Packard Co
+   Copyright (C) 2001-2004 Hewlett-Packard Co
 	Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
 
 This file is part of libunwind.
@@ -31,16 +31,16 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 
 #include <libunwind-ia64.h>
 
-#include "config.h"
-#include "internal.h"
-#include "rse.h"
+#include "ia64/rse.h"
 
-#define IA64_UNW_VER(x)			((x) >> 48)
-#define IA64_UNW_FLAG_MASK		0x0000ffff00000000
-#define IA64_UNW_FLAG_OSMASK		0x0000f00000000000
-#define IA64_UNW_FLAG_EHANDLER(x)	((x) & 0x0000000100000000L)
-#define IA64_UNW_FLAG_UHANDLER(x)	((x) & 0x0000000200000000L)
-#define IA64_UNW_LENGTH(x)		((x) & 0x00000000ffffffffL)
+#include "internal.h"
+
+#define IA64_UNW_VER(x)		    ((x) >> 48)
+#define IA64_UNW_FLAG_MASK	    ((unw_word_t) 0x0000ffff00000000ULL)
+#define IA64_UNW_FLAG_OSMASK	    ((unw_word_t) 0x0000f00000000000ULL)
+#define IA64_UNW_FLAG_EHANDLER(x)   ((x) & (unw_word_t) 0x0000000100000000ULL)
+#define IA64_UNW_FLAG_UHANDLER(x)   ((x) & (unw_word_t) 0x0000000200000000ULL)
+#define IA64_UNW_LENGTH(x)	    ((x) & (unw_word_t) 0x00000000ffffffffULL)
 
 #ifdef MIN
 # undef MIN
@@ -49,7 +49,89 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 
 #include "tdep.h"
 
-/* Bits 0 and 1 of an location are used to encode its type:
+#if !defined(HAVE_SYS_UC_ACCESS_H) && !defined(UNW_REMOTE_ONLY)
+
+static inline ALWAYS_INLINE void *
+inlined_uc_addr (ucontext_t *uc, int reg, uint8_t *nat_bitnr)
+{
+  unw_word_t reg_addr;
+  void *addr;
+
+  switch (reg)
+    {
+    case UNW_IA64_GR + 0:	addr = &unw.read_only.r0; break;
+    case UNW_IA64_NAT + 0:	addr = &unw.read_only.r0; break;
+    case UNW_IA64_FR + 0:	addr = &unw.read_only.f0; break;
+    case UNW_IA64_FR + 1:
+      if (__BYTE_ORDER == __BIG_ENDIAN)
+	addr = &unw.read_only.f1_be;
+      else
+	addr = &unw.read_only.f1_le;
+      break;
+    case UNW_IA64_IP:		addr = &uc->uc_mcontext.sc_br[0]; break;
+    case UNW_IA64_CFM:		addr = &uc->uc_mcontext.sc_ar_pfs; break;
+    case UNW_IA64_AR_RNAT:	addr = &uc->uc_mcontext.sc_ar_rnat; break;
+    case UNW_IA64_AR_UNAT:	addr = &uc->uc_mcontext.sc_ar_unat; break;
+    case UNW_IA64_AR_LC:	addr = &uc->uc_mcontext.sc_ar_lc; break;
+    case UNW_IA64_AR_FPSR:	addr = &uc->uc_mcontext.sc_ar_fpsr; break;
+    case UNW_IA64_PR:		addr = &uc->uc_mcontext.sc_pr; break;
+      /* This may look confusing, but it's correct: AR_BSPSTORE needs
+         to return the address past the last word written, which is
+         stored in sc_ar_bsp.  On the other hand, AR_BSP needs to
+         return the address that was in ar.bsp at the time the context
+         was captured.  As described in unw_init_local(), sc_ar_bsp is
+         (ab-)used for this purpose.  */
+    case UNW_IA64_AR_BSP:	addr = &uc->uc_mcontext.sc_rbs_base; break;
+    case UNW_IA64_AR_BSPSTORE:	addr = &uc->uc_mcontext.sc_ar_bsp; break;
+
+    case UNW_IA64_GR + 4 ... UNW_IA64_GR + 7:
+    case UNW_IA64_GR + 12:
+      addr = &uc->uc_mcontext.sc_gr[reg - UNW_IA64_GR];
+      break;
+
+    case UNW_IA64_NAT + 4 ... UNW_IA64_NAT + 7:
+    case UNW_IA64_NAT + 12:
+      addr = &uc->uc_mcontext.sc_nat;
+      reg_addr = (unw_word_t) &uc->uc_mcontext.sc_gr[reg - UNW_IA64_NAT];
+      *nat_bitnr = reg - UNW_IA64_NAT;
+      break;
+
+    case UNW_IA64_BR + 1 ... UNW_IA64_BR + 5:
+      addr = &uc->uc_mcontext.sc_br[reg - UNW_IA64_BR];
+      break;
+
+    case UNW_IA64_FR+ 2 ... UNW_IA64_FR+ 5:
+    case UNW_IA64_FR+16 ... UNW_IA64_FR+31:
+      addr = &uc->uc_mcontext.sc_fr[reg - UNW_IA64_FR];
+      break;
+
+    default:
+      addr = NULL;
+    }
+  return addr;
+}
+
+static inline void *
+uc_addr (ucontext_t *uc, int reg, uint8_t *nat_bitnr)
+{
+  if (__builtin_constant_p (reg))
+    return inlined_uc_addr (uc, reg, nat_bitnr);
+  else
+    return tdep_uc_addr (uc, reg, nat_bitnr);
+}
+
+/* Return TRUE if ADDR points inside unw.read_only_reg.  */
+
+static inline long
+ia64_read_only_reg (void *addr)
+{
+  return ((unsigned long) ((char *) addr - (char *) &unw.read_only)
+	  < sizeof (unw.read_only));
+}
+
+#endif /* !defined(HAVE_SYS_UC_ACCESS_H) && !defined(UNW_REMOTE_ONLY) */
+
+/* Bits 0 and 1 of a location are used to encode its type:
 	bit 0: set if location uses floating-point format.
 	bit 1: set if location is a NaT bit on memory stack.  */
 
@@ -57,11 +139,9 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #define IA64_LOC_TYPE_MEMSTK_NAT	(1 << 1)
 
 #ifdef UNW_LOCAL_ONLY
-
 #define IA64_LOC_REG(r,t)	(((r) << 2) | (t))
 #define IA64_LOC_ADDR(a,t)	(((a) & ~0x3) | (t))
 #define IA64_LOC_UC_ADDR(a,t)	IA64_LOC_ADDR(a, t)
-#define IA64_LOC_UC_REG(a,t)
 #define IA64_NULL_LOC		(0)
 
 #define IA64_GET_REG(l)		((l) >> 2)
@@ -72,9 +152,10 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #define IA64_IS_REG_LOC(l)	0
 #define IA64_IS_UC_LOC(l)	0
 
-#define IA64_REG_LOC(c,r)	((unw_word_t) tdep_uc_addr((c)->as_arg, (r)))
+#define IA64_REG_LOC(c,r)	((unw_word_t) uc_addr((c)->as_arg, r, NULL))
+#define IA64_REG_NAT_LOC(c,r,n)	((unw_word_t) uc_addr((c)->as_arg, r, n))
 #define IA64_FPREG_LOC(c,r)						 \
-	((unw_word_t) tdep_uc_addr((c)->as_arg, (r)) | IA64_LOC_TYPE_FP)
+	((unw_word_t) uc_addr((c)->as_arg, (r), NULL) | IA64_LOC_TYPE_FP)
 
 # define ia64_find_proc_info(c,ip,n)					\
 	tdep_find_proc_info(unw_local_addr_space, (ip), &(c)->pi, (n),	\
@@ -82,15 +163,15 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 # define ia64_put_unwind_info(c, pi)	do { ; } while (0)
 
 /* Note: the register accessors (ia64_{get,set}{,fp}()) must check for
-   NULL locations because tdep_uc_addr() returns NULL for unsaved
+   NULL locations because uc_addr() returns NULL for unsaved
    registers.  */
 
 static inline int
 ia64_getfp (struct cursor *c, unw_word_t loc, unw_fpreg_t *val)
 {
-  if (!loc)
+  if (IA64_IS_NULL_LOC (loc))
     {
-      debug (150, "%s: access to unsaved register\n", __FUNCTION__);
+      Debug (16, "access to unsaved register\n");
       return -UNW_EBADREG;
     }
   *val = *(unw_fpreg_t *) IA64_GET_ADDR (loc);
@@ -100,21 +181,28 @@ ia64_getfp (struct cursor *c, unw_word_t loc, unw_fpreg_t *val)
 static inline int
 ia64_putfp (struct cursor *c, unw_word_t loc, unw_fpreg_t val)
 {
-  if (!loc)
+  unw_fpreg_t *addr = (unw_fpreg_t *) IA64_GET_ADDR (loc);
+
+  if (IA64_IS_NULL_LOC (loc))
     {
-      debug (150, "%s: access to unsaved register\n", __FUNCTION__);
+      Debug (16, "access to unsaved register\n");
       return -UNW_EBADREG;
     }
-  *(unw_fpreg_t *) IA64_GET_ADDR (loc) = val;
+  else if (ia64_read_only_reg (addr))
+    {
+      Debug (16, "attempt to read-only register\n");
+      return -UNW_EREADONLYREG;
+    }
+  *addr = val;
   return 0;
 }
 
 static inline int
 ia64_get (struct cursor *c, unw_word_t loc, unw_word_t *val)
 {
-  if (!loc)
+  if (IA64_IS_NULL_LOC (loc))
     {
-      debug (150, "%s: access to unsaved register\n", __FUNCTION__);
+      Debug (16, "access to unsaved register\n");
       return -UNW_EBADREG;
     }
   *val = *(unw_word_t *) IA64_GET_ADDR (loc);
@@ -124,12 +212,19 @@ ia64_get (struct cursor *c, unw_word_t loc, unw_word_t *val)
 static inline int
 ia64_put (struct cursor *c, unw_word_t loc, unw_word_t val)
 {
-  if (!loc)
+  unw_word_t *addr = (unw_word_t *) IA64_GET_ADDR (loc);
+
+  if (IA64_IS_NULL_LOC (loc))
     {
-      debug (150, "%s: access to unsaved register\n", __FUNCTION__);
+      Debug (16, "access to unsaved register\n");
       return -UNW_EBADREG;
     }
-  *(unw_word_t *) IA64_GET_ADDR (loc) = (val);
+  else if (ia64_read_only_reg (addr))
+    {
+      Debug (16, "attempt to read-only register\n");
+      return -UNW_EREADONLYREG;
+    }
+  *addr = val;
   return 0;
 }
 
@@ -163,6 +258,7 @@ ia64_put (struct cursor *c, unw_word_t loc, unw_word_t val)
 #define IA64_IS_UC_LOC(l)	(((l).w1 & IA64_LOC_TYPE_UC) != 0)
 
 #define IA64_REG_LOC(c,r)	IA64_LOC_REG ((r), 0)
+#define IA64_REG_NAT_LOC(c,r,n)	IA64_LOC_REG ((r), 0)
 #define IA64_FPREG_LOC(c,r)	IA64_LOC_REG ((r), IA64_LOC_TYPE_FP)
 
 # define ia64_find_proc_info(c,ip,n)					\
@@ -185,6 +281,12 @@ ia64_getfp (struct cursor *c, ia64_loc_t loc, unw_fpreg_t *val)
   unw_word_t addr;
   int ret;
 
+  if (IA64_IS_NULL_LOC (loc))
+    {
+      Debug (16, "access to unsaved register\n");
+      return -UNW_EBADREG;
+    }
+
   if (IA64_IS_UC_LOC (loc))
     return ia64_uc_access_fpreg (c, loc, val, 0);
 
@@ -208,6 +310,12 @@ ia64_putfp (struct cursor *c, ia64_loc_t loc, unw_fpreg_t val)
   unw_word_t addr;
   int ret;
 
+  if (IA64_IS_NULL_LOC (loc))
+    {
+      Debug (16, "access to unsaved register\n");
+      return -UNW_EBADREG;
+    }
+
   if (IA64_IS_UC_LOC (loc))
     return ia64_uc_access_fpreg (c, loc, &val, 1);
 
@@ -233,6 +341,12 @@ ia64_putfp (struct cursor *c, ia64_loc_t loc, unw_fpreg_t val)
 static inline int
 ia64_get (struct cursor *c, ia64_loc_t loc, unw_word_t *val)
 {
+  if (IA64_IS_NULL_LOC (loc))
+    {
+      Debug (16, "access to unsaved register\n");
+      return -UNW_EBADREG;
+    }
+
   if (IA64_IS_FP_LOC (loc))
     {
       unw_fpreg_t tmp;
@@ -263,6 +377,12 @@ ia64_get (struct cursor *c, ia64_loc_t loc, unw_word_t *val)
 static inline int
 ia64_put (struct cursor *c, ia64_loc_t loc, unw_word_t val)
 {
+  if (IA64_IS_NULL_LOC (loc))
+    {
+      Debug (16, "access to unsaved register\n");
+      return -UNW_EBADREG;
+    }
+
   if (IA64_IS_FP_LOC (loc))
     {
       unw_fpreg_t tmp;
@@ -369,9 +489,9 @@ struct ia64_labeled_state
 #define ia64_create_state_record	UNW_OBJ(create_state_record)
 #define ia64_free_state_record		UNW_OBJ(free_state_record)
 #define ia64_find_save_locs		UNW_OBJ(find_save_locs)
+#define ia64_validate_cache		UNW_OBJ(ia64_validate_cache)
+#define ia64_local_validate_cache	UNW_OBJ(ia64_local_validate_cache)
 #define ia64_per_thread_cache		UNW_OBJ(per_thread_cache)
-#define ia64_access_reg			UNW_OBJ(access_reg)
-#define ia64_access_fpreg		UNW_OBJ(access_fpreg)
 #define ia64_scratch_loc		UNW_OBJ(scratch_loc)
 #define ia64_local_resume		UNW_OBJ(local_resume)
 #define ia64_local_addr_space_init	UNW_OBJ(local_addr_space_init)
@@ -380,7 +500,6 @@ struct ia64_labeled_state
 #define rbs_switch			UNW_OBJ(rbs_switch)
 #define rbs_find_stacked		UNW_OBJ(rbs_find_stacked)
 #define rbs_cover_and_flush		UNW_OBJ(rbs_cover_and_flush)
-#define ia64_init			UNW_ARCH_OBJ(init)
 
 extern int ia64_make_proc_info (struct cursor *c);
 extern int ia64_fetch_proc_info (struct cursor *c, unw_word_t ip,
@@ -391,17 +510,15 @@ extern int ia64_create_state_record (struct cursor *c,
 				     struct ia64_state_record *sr);
 extern int ia64_free_state_record (struct ia64_state_record *sr);
 extern int ia64_find_save_locs (struct cursor *c);
+extern void ia64_validate_cache (unw_addr_space_t as, void *arg);
+extern int ia64_local_validate_cache (unw_addr_space_t as, void *arg);
 extern void ia64_local_addr_space_init (void);
-extern void ia64_init (void);
-extern int ia64_access_reg (struct cursor *c, unw_regnum_t reg,
-			    unw_word_t *valp, int write);
-extern int ia64_access_fpreg (struct cursor *c, unw_regnum_t reg,
-			      unw_fpreg_t *valp, int write);
-extern ia64_loc_t ia64_scratch_loc (struct cursor *c, unw_regnum_t reg);
+extern ia64_loc_t ia64_scratch_loc (struct cursor *c, unw_regnum_t reg,
+				    uint8_t *nat_bitnr);
 
-extern void ia64_install_cursor (struct cursor *c, unw_word_t pri_unat,
-				 unw_word_t *extra)
-	__attribute__ ((noreturn));
+extern NORETURN void ia64_install_cursor (struct cursor *c,
+					  unw_word_t pri_unat,
+					  unw_word_t *extra);
 extern int ia64_local_resume (unw_addr_space_t as, unw_cursor_t *cursor,
 			      void *arg);
 extern int rbs_switch (struct cursor *c,
@@ -434,7 +551,7 @@ rbs_contains (struct rbs_area *rbs, unw_word_t bsp)
      test is (bsp >= rbs->end - rbs->size) && (bsp < rbs->end).  We
      take advantage of the fact that -n == ~n + 1.  */
   result = bsp - rbs->end > ~rbs->size;
-  debug (150, "%s: 0x%lx in [0x%lx-0x%lx) => %d\n", __FUNCTION__,
+  Debug (16, "0x%lx in [0x%lx-0x%lx) => %d\n",
 	 (long) bsp, (long) (rbs->end - rbs->size), (long) rbs->end, result);
   return result;
 }
@@ -487,6 +604,11 @@ ia64_get_stacked (struct cursor *c, unw_word_t reg,
   return ret;
 }
 
+/* The UNaT slot # calculation is identical to the one for RNaT slots,
+   but for readability/clarity, we don't want to use
+   ia64_rnat_slot_num() directly.  */
+#define ia64_unat_slot_num(addr)	ia64_rse_slot_num(addr)
+
 /* XXX should be in glibc: */
 #ifndef IA64_SC_FLAG_ONSTACK
 # define IA64_SC_FLAG_ONSTACK_BIT    0 /* running on signal stack? */
